/*
 * module.hpp
 *
 *  Created on: 20/02/2013
 *      Author: mmcgill
 */

#ifndef CROSBOT_FASTSLAM_MODULE_HPP_
#define CROSBOT_FASTSLAM_MODULE_HPP_

#include <ros/ros.h>

#include <crosbot/controls/command.hpp>
#include <crosbot_fastslam/fastslam.hpp>
#include <geometry_msgs/Pose.h>
#include <nav_msgs/Path.h>
#include <sensor_msgs/LaserScan.h>
#include <tf/transform_listener.h>
#include <tf/transform_broadcaster.h>

namespace crosbot {

namespace fastslam {

#define DEFAULT_ODOMFRAME			"/icp"
#define DEFAULT_MAPFRAME			"/map"
#define DEFAULT_BASEFRAME			"/base_link"
#define DEFAULT_MAXWAIT4TRANSFORM	2.0			// [s]

class FastSLAMModule;
typedef Handle<FastSLAMModule> FastSLAMModulePtr;

class FastSLAMModule : public FastSLAMMap, virtual public MapListener {
protected:
	// ROS Configuration
    std::string baseFrame;
    std::string mapFrame;
	std::string odomFrame;
	std::string pub_historyName;
	std::string pub_mapName;
	std::string pub_particlesName;
	std::string srv_mapServerName;
	std::string sub_poseEstimate;
	std::string sub_scanName;
	std::string sub_snapName;

	ros::ServiceClient mapServerSrv;
	ros::Subscriber poseEstimateSub;
	ros::Subscriber scanSub;
	ros::Subscriber snapSub;

	ros::Publisher gridPub;
	ros::Publisher historyPub;
	ros::Publisher particlesPub;

	crosbot::CrosbotCommandPtr crosbotCommand;

	tf::TransformListener tfListener;
    tf::TransformBroadcaster tfBroadcast;
public:

    FastSLAMModule(std::string name = "");
    virtual ~FastSLAMModule();

    void initialize(ros::NodeHandle& nh);
    void shutdown();

	// ROS Callbacks
    void callbackPoseEstimate(const geometry_msgs::PointStampedConstPtr& poseEstimate);
    void callbackScan(const sensor_msgs::LaserScanConstPtr& latestScan);
	void callbackSnap(const crosbot_map::SnapMsgConstPtr& snap);

	void publishTransform(const Pose& robot, const Pose& odometry, const Time& time);

	// Map Listener callbacks
	void motionTracked(MapPtr map);
	void mapUpdated(MapPtr map);
	void tagAdded(MapPtr map, TagPtr tag);
	void tagChanged(MapPtr map, TagPtr tag);

	// Crosbot command callback
	void callback_receivedCommand(crosbot_msgs::ControlCommandPtr command);

protected:

	virtual void resetFastSlam();

private:

	/**
	 * Estimates the pose of the robot within the entire map.
	 * Does not work well.
	 */
	void estimatePoseInMap()  __attribute__ ((deprecated));
};

} // namespace fastslam

} // namespace crosbot

#endif /* CROSBOT_FASTSLAM_MODULE_HPP_ */
